package prolog;

import prolog.util.Vector;

/* The JavaClass package. */
import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;

/**
 * User-specified algebraic data types. We assume that each component has a
 * distinct functor/arity.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: AlgebraicType.java,v 1.7 2000/11/13 11:55:03 peteg Exp $
 * @see prolog.lib._type_def_2
 */
class AlgebraicType extends TermType {
    private String name;
    private int functor_arity;

    int numVars;
    CompoundType[] components;

    AlgebraicType(String name, int functor_arity, int numVars) {
	this(name, functor_arity, numVars, null);
    }

    // Used by unify()
    private AlgebraicType(String name,
			  int functor_arity,
			  int numVars,
			  CompoundType[] components) {
	this.name          = name;
	this.functor_arity = functor_arity;
	this.numVars       = numVars;
	this.components    = components;
    }

    void setComponents(Vector components) {
	this.components = new CompoundType[components.size()];

	for(int i = components.size(); i-- > 0; ) {
	    this.components[i] = (CompoundType)components.elementAt(i);
	}
    }

    ObjectType getJavaType(TermInfo ti) {
	return CompoundType.JavaType;
    }

    TermInfo instantiate(Term t, TermInfo[] typevars,
			 TypeDictionary types, boolean map) {
// 	System.err.println("AlgebraicType.instantiate(): " + this + " against " + t);

	Compound c = (Compound)t.getBinding();

	if(c.ti == null) {
	    return findConstructor(c.functor_arity).instantiate(c,
								typevars,
								types,
								map);
	} else {
	    return c.ti;
	}
    }

    boolean unify(TermInfo ti,
		  TermInfo arg,
		  TypeTrail trail) {
// 	System.err.println("AlgebraicType.unify(): " + this + " against " + arg.type);

 	if(arg.type instanceof CompoundType) {
	    // Unify with a component of this type.
	    return findConstructor(((CompoundType)arg.type).functor_arity).unify(ti,
										 arg,
										 trail);
	}

	if(arg.type instanceof AlgebraicType && arg.type == this) {
	    // Consider the arg instantiations only.
	    for(int i = 0; i < numVars; i++) {
		TermInfo this_var_ti = ti.typevars[i];
		TermInfo arg_var_ti  = arg.typevars[i];

// 		System.err.println(" >> " + this_var_ti
// 				   + " <-> "
// 				   + arg_var_ti);

		if(this_var_ti == null) {
		    // We'll unify with anything.
		    ti.typevars[i] = arg_var_ti;
		} else {
		    if(arg_var_ti == null) {
			// It'll unify with anything.
			arg.typevars[i] = this_var_ti;
		    } else {
			// Actually unify the two...
			if(!this_var_ti.type.unify(this_var_ti,
						   arg_var_ti,
						   trail)) {
			    return false;
			}
		    }
		}
	    }

	    return true;
	}

	// Nope.
	return false;
    }

    public boolean compile_unify(ClauseCodeGen ccg,
				 TermInfo arg0, TermInfo arg1) {
 	if(arg1.type instanceof TypeVariable) {
 	    // Reverse data flow.
 	    return arg1.type.compile_unify(ccg, arg1, arg0);
 	}

//   	System.err.println("AlgebraicType.compile_unify(): "
// 			   + this
// 			   + " <-> "
// 			   + arg1.type);

	if(arg1.location == null) {
// 	    System.err.println(" >> destructor.");
	    // Destructor -- check the typing.
	    CompoundType ct = findConstructor(((CompoundType)arg1.type).functor_arity);

	    if(ct == null) {
		// Ill-typed.
		throw new RuntimeException("AlgebraicType.compile_unify(): ill typed!");
	    } else {
		// Well-typed.
		ccg.compound_checkFunctor(arg0, ct.functor_arity);

		// The type of this term is now the specific component.
// 		System.err.println(" >> AlgebraicType.compile_unify(): -> " + ct);
		arg0.type = ct;

		// Add some ti's for imaginary arg's, located relative to
		// this thing.
		TermInfo[] arg_tis = arg0.arg_tis = new TermInfo[ct.arity];

		for(int i = 0; i < ct.arity; i++) {
		    arg_tis[i] = new TermInfo(ct.args[i], arg0.typevars,
					      new Location(arg0, i));
		}

		return ct.compile_unify(ccg, arg0, arg1);
	    }
	} else {
	    // Shouldn't happen.
	    throw new RuntimeException("AlgebraicType.compile_unify()");
	}
    }

    public void create(ClauseCodeGen ccg, TermInfo arg0) {
	throw new RuntimeException();
    }

    /**
     * Finds the constructor associated with a given functor/arity.
     *
     * @param update become the found sub-component.
     */
    private CompoundType findConstructor(int fa) {
	for(int i = 0; i < components.length; i++) {
	    if(components[i].functor_arity == fa) {
		return components[i];
	    }
	}

	return null;
    }

    /**
     * Type + Definition.
     */
    public void toStringDef(StringBuffer out) {
	out.append(toString());

	if(components != null) {
	    for(int i = 0; i < components.length; i++) {
		out.append("\n  ");
		components[i].toString(out);
	    }
	}
    }

    /**
     * Type only.
     */
    public void toString(StringBuffer out) {
	out.append("(alg) ");
	out.append(name);
    }
}
