package prolog;

import java.io.*;

import prolog.lib.unify_2;
import prolog.util.Hashtable;
import prolog.util.Vector;

/**
 * Represents a clause, which consist of a head (a <TT>Compound</TT> Term)
 * and, optionally, a body (a <TT>GoalList</TT>).
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Clause.java,v 1.17 2000/11/12 12:22:44 peteg Exp $
 */
public class Clause implements Cloneable, Serializable {
    public Compound head;
    public Goal body;

    /**
     * The parsed version of this clause.
     */
    public Compound parsed;

    /**
     * Number of vars in this clause.
     */
    public int numVars;

    /**
     * Creates a new <TT>Clause</TT>, converting a parsed clause of either
     * of the forms: <TT>Head.</TT> or <TT>':-'(Head, Body)</TT> into the
     * internal representation.
     *
     * @param parsed the parsed form of this clause.
     */
    public Clause(Compound parsed, Prolog p) {
	Hashtable vars = new Hashtable();

	this.parsed  = parsed = parsed.squash(vars);
	this.numVars = vars.size();

	if(parsed.functor == Parser.NECK_FUNCTOR
	   && parsed.arity == Parser.NECK_ARITY) {
	    head = (Compound)parsed.args[0];
	    body = p.toGoal((Compound)parsed.args[1]);
	} else {
	    head = parsed;
	    body = null;
	}
    }

    /**
     * Creates a new <TT>Clause</TT>.
     *
     * @param head the head of this clause.
     * @param body the body of this clause.
     * @param parsed the parsed form of this clause: <TT>':-'(Head, Body)</TT>.
     * @param numVars the number of variables in this clause.
     */
    public Clause(Compound head, Goal body,
		  Compound parsed, int numVars) {
	this.head    = head;
	this.body    = body;
	this.parsed  = parsed;
	this.numVars = numVars;
    }

    /**
     * Normalises this clause into the format specified in andrewt's
     * doctoral thesis (more or less). This makes it easier to compile as
     * unifications are tightly restricted in form.
     * <P>
     * Note we don't <TT>typeify()</TT> the top-level arguments - so if
     * <TT>head_args[i].ti == null</TT>, arg i isn't used.
     *
     * @see prolog.ClauseIndex#compileClauses
     * @see prolog.Goal#normalise
     */
    Clause normalise(TypeDictionary types) {
	// Assume we're not going to introduce too many variables.
	Vector new_vars = new Vector(10);

	// Maps old ids -> new vars.
	Variable[] vars = new Variable[numVars];
	int next_new_var = numVars;

	/* Head transform: replace successive occurrences of variables
	 * and non-variables with new ones (i.e. linearise). Add
	 * unify/2's to the body to preserve meaning. */
	Term[] new_head_args = new Term[head.arity];
	Goal new_body = body;

	Conjunction ptr = null;

	for(int i = 0; i < head.arity; i++) {
	    Term arg = head.args[i];

	    Variable var = new Variable(next_new_var++);
	    new_head_args[i] = var;
	    new_vars.addElement(var);
	    var.typify(types, false);

	    if(arg instanceof Variable) {
		Variable v = (Variable)arg;

		if(vars[v.id] == null) {
		    vars[v.id] = var;
		    continue;
		}
	    }

	    Conjunction unify = new Conjunction(new Compound(unify_2.functor,
							     new Term[]{var, arg}),
						unify_2.unify);

	    if(ptr == null) {
		unify.next = new_body;
		new_body = ptr = unify;
	    } else {
		unify.next = ptr.next;
		ptr.next = unify;
		ptr = unify;
	    }
	}

	// Now normalise the body.
	new_body = new_body.normalise(vars, next_new_var, new_vars, types);

// 	System.err.println(new_body);

	// Renumber the variables -- not strictly necessary...
	// All typevar's attached to variables share an environment.
	TermInfo[] global_environment = new TermInfo[new_vars.size()];

	for(int i = new_vars.size(); i-- > 0; ) {
	    Variable var = (Variable)new_vars.elementAt(i);

// 	    System.err.print("Renaming var: " + var +  " -> ");

	    int newID = var.getID() - numVars;

	    var.setID(newID);

	    // Anonymous variables won't be given a type.
// 	    if(var.ti != null) {
		((TypeVariable)var.ti.type).setID(newID);
		var.ti.typevars = global_environment;
// 	    }

//  	    System.err.println(var);
	}

	Clause new_clause = new Clause(new Compound(head.functor,
						    new_head_args),
				       new_body,
				       parsed, new_vars.size());

//  	System.err.println(">> Clause.normalise():\n" + new_clause);

	return new_clause;
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append(head.toString());

	if(body != null) {
	    out.append(" :-\n\t");
	    out.append(body.toString());
	}

	out.append(".\n");

	return out.toString();
    }
}
