package prolog;

/* The JavaClass package. */
import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;

import java.util.Enumeration;

import prolog.math.*;
import prolog.util.Hashtable;
import prolog.util.Vector;

/**
 * Represents a point we need to patch to branch to failure code.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: ClauseCodeGen.java,v 1.32 2000/11/16 10:30:07 peteg Exp $
 */
class FailurePoint {
    BranchInstruction fp;
    int bindings;

    FailurePoint next;

    FailurePoint(BranchInstruction bi, int bindings) {
	fp = bi;
	this.bindings = bindings;
    }

    public String toString() {
	return "FailurePoint: fp = " + fp + " bindings_index = " + bindings;
    }
}

/**
 * Generates code for a single clause. Provides macros for code generation,
 * and is the repository for fragments of generated code.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: ClauseCodeGen.java,v 1.32 2000/11/16 10:30:07 peteg Exp $
 * @see prolog.PredicateCodeGen
 */
public class ClauseCodeGen {
    ClauseCodeGen next;

    private InstructionList il;
    private ConstantPoolGen cp;
    private InstructionFactory inst_fact;

    String functor;
    int arity;

    public TypeDictionary types;

    /**
     * Index of the first available local variable.
     */
    private short first_avail_lv = JITDefs.first_free_lv;

    /**
     * Points at which to patch branch-to-failure-code-if-false
     * into. (linked list stack).
     */
    private FailurePoint failure_points;
 
    /**
     * Does this clause contain calls to predicates with side-effects? If
     * so, we need to retain it, even if it fails.
     */
    boolean sideEffects = false;

    /**
     * Can this clause fail?
     */
    boolean canFail;

    String class_name;

    /**
     * The set of predicates we invoke.
     */
    Hashtable predicates;

    Hashtable functors;

    TypeTrail trail;

    Vector bindings;

    /**
     * Creates a new compilation context for a clause.
     *
     * @param tid the parent (predicate-level) namespace.
     */
    ClauseCodeGen(PredicateCodeGen pcg) {
	this.cp        = pcg.cp;
	this.inst_fact = pcg.inst_fact;
	this.il        = pcg.il;

	this.types     = pcg.types;

	this.functor   = pcg.functor;
	this.arity     = pcg.arity;

	this.class_name = pcg.class_name;

	this.predicates = pcg.predicates;
	this.functors   = pcg.functors;

	this.trail      = new TypeTrail();
	this.bindings   = new Vector();
    }

    /**
     * Compiles a clause. Assumes the clause has been normalised.
     *
     * @param ci_det this clause is in a determinate situ according to the
     * indexer.
     * @return <TT>false</TT> if this clause fails with no side effects,
     * <TT>true</TT> otherwise.
     */
    boolean compile(Clause clause, TermInfo[] type, boolean ci_det) {
	clause = clause.normalise(types);

// 	System.err.println(">> ClauseCodeGen.compile(" + ci_det + "):\n" + clause);

	/* Bind each argument in the head to it's type. */
	Compound head         = clause.head;
	Term[] head_args      = head.args;

	TermInfo args_loc = new TermInfo(null);
	args_loc.location = new Location(null, JITDefs.args_lv);

 	for(int i = 0; i < head.arity; i++) {
	    TermInfo ti0 = head_args[i].ti;
	    TermInfo ti1 = type[i];

	    // Unravel ti1's typevar chains.
//  	    ti1 = ti1.type.getBinding(ti1);

 	    ti0.set(ti1, new Location(args_loc, i));
 	}

	if(clause.body.compile(this)) {
	    patchFP();

	    // HACK: reset all ti's in the original type.
	    for(int i = 0; i < type.length; i++) {
		type[i].clean();
	    }

	    return true;
	} else {
	    throw new RuntimeException();
// 	    return false;
	}
    }

    /**
     * Patches the failure points to branch to the end of the clause.
     */
    void patchFP() {
	// If we fall off the end of the list of goals, we've succeeded.
	il.append(inst_fact.ICONST_1);
	il.append(inst_fact.IRETURN);
	il.append(inst_fact.NOP); // FIXME: indexer screwiness.

	/* Return false on failure. Variables are unbound by the backtrack
         * code in the interpreter. FIXME: this isn't always true if there's
         * backtracking within a compiled predicate, i.e. shallow
         * backtracking. */
	if(failure_points != null) {
	    InstructionHandle ih = il.append(inst_fact.NOP);
	    int bindings_index = 0;

	    while(failure_points != null) {
// 		System.err.println(">> patch fp: " + failure_points);

		failure_points.fp.setTarget(ih);

		while(bindings_index < failure_points.bindings) {
		    // FIXME: unbind bindings here.
		    TermInfo ti = (TermInfo)bindings.elementAt(bindings_index);

// 		    System.err.println(">> adding unbind code for: " + ti);

		    ti.load(this, false); // Shouldn't have to cast it.
		    il.append(inst_fact.createInvoke(TypeVariable.ClassName,
						     JITDefs.binding_unbind,
						     JITDefs.binding_unbind_ret_type,
						     JITDefs.binding_unbind_args_type,
						     JITDefs.INVOKEVIRTUAL));
		    bindings_index++;
		}

		failure_points = failure_points.next;
	    }

	}
    }

    /**
     * Compiles a generic call to a predicate.
     */
    public boolean compile_call(Predicate predicate, Term[] args,
				boolean lastcall) {

	String functor = predicate.functor;
	int arity = predicate.arity;

// 	System.err.println("CCG.call(): " + functor + "/" + arity
// 			   + " lastcall: " + lastcall);

// 	for(int i = 0; i < arity; i++) {
// 	    Term arg = args[i];

// 	    System.err.println("   " + arg
// 			       + "   " + arg.ti);
// 	}

	if(lastcall && class_name.equals(JITDefs.package_name + "."
					 + functor + "_" + arity)) {
// 	    System.err.println(" >> TRO.");

	    // Last call. If the name matches that of this predicate, update
	    // args and branch back to the start.

	    // FIXME: types should be checked.

	    il.append(JITDefs.args_load_inst);

	    for(int i = 0; i < arity; i++) {
		if(i < arity - 1) {
		    il.append(inst_fact.DUP);
		}

		il.append(new PUSH(cp, i));
		args[i].ti.load(this, false);
		il.append(inst_fact.AASTORE);
	    }

	    il.append(new GOTO(il.getStart()));

	    return true;
	} else {
	    // General version.
	    String static_field_name = functor + "_" + arity;

	    predicates.put(static_field_name, predicate);

	    il.append(inst_fact.createGetStatic(class_name,
						static_field_name,
						JITDefs.predicate_type));

	    if(arity > 0) {
		il.append(new PUSH(cp, arity));
		il.append(new ANEWARRAY(cp.addClass(TermType.JavaType)));

		for(int i = 0; i < arity; i++) {
		    il.append(inst_fact.DUP);
		    il.append(new PUSH(cp, i));
		    args[i].ti.load(this, false);
		    il.append(inst_fact.AASTORE);
		}
	    } else {
		il.append(inst_fact.createNull(JITDefs.term_array_type));
	    }

	    il.append(JITDefs.prolog_load_inst);
	    il.append(inst_fact.createInvoke(JITDefs.predicate_class_name,
					     JITDefs.call_name,
					     JITDefs.call_ret_type,
					     JITDefs.call_arg_type,
					     JITDefs.INVOKEVIRTUAL));

	    addFP(ifeq());
	}

	// We don't know anything about how the predicate will behave,
	// except that we assume it can only bind Bindings (i.e. no setarg/3
	// stuff!). Change the arg types to reflect this.
// 	for(int i = 0; i < arity; i++) {
// 	    args[i].set_type(tid, TermType.type);
// 	}

	return true;
    }

    /**
     * Adds a failure point to the list. It is assumed that there is a
     * boolean value on the stack, with true == success, continue, false ==
     * branch to failure code.
     */
    public void addFP(BranchInstruction bi) {
	FailurePoint fp;

	fp = new FailurePoint(bi, bindings.size());
	fp.next = failure_points;
	failure_points = fp;

	canFail = true;
    }

    public void checkcast(ObjectType t) {
	if(t == TermType.JavaType) {
	    throw new RuntimeException();
	}

	il.append(inst_fact.createCheckCast(t));
    }

    public void dup() {
	il.append(inst_fact.DUP);
    }

    public void dup_x2() {
	il.append(inst_fact.DUP_X2);
    }

    public BranchInstruction ifeq() {
	BranchInstruction bi;

	il.append(bi = new IFEQ(null));
	return bi;
    }

    public BranchInstruction ifne() {
	BranchInstruction bi;

	il.append(bi = new IFNE(null));
	return bi;
    }

    public BranchInstruction if_icmpne() {
	BranchInstruction bi;

	il.append(bi = new IF_ICMPNE(null));
	return bi;
    }

    public BranchInstruction if_acmpne() {
	BranchInstruction bi;

 	il.append(bi = new IF_ACMPNE(null));
	return bi;
    }

    public void pop() {
	il.append(inst_fact.POP);
    }

    /**
     * Allocates a local variable to a <TT>Term</TT>.
     */
    public LocalVariableGen lv_alloc(Type type) {
	int lv_num = first_avail_lv++;

	return new LocalVariableGen(lv_num,
				    "lv" + lv_num,
				    type,
				    null,
				    null);
    }

    public void lv_load(LocalVariableGen lv) {
	il.append(new ALOAD(lv.getIndex()));
    }

    public void lv_store(LocalVariableGen lv) {
	il.append(new ASTORE(lv.getIndex()));
    }

    public InstructionHandle nop() {
	return il.append(inst_fact.NOP);
    }

    /**
     * Load an an element of a <TT>Term[]</TT> array (reference to it on top
     * of the stack).
     * <P>
     * Stack: <TT>..., Term[] Ref -> ..., Term</TT>
     */
    void arg_load(short index) {
	il.append(new PUSH(cp, index));
	il.append(inst_fact.createArrayLoad(TermType.JavaType));
    }

    /**
     * Loads the <TT>arrayref</TT> and <TT>index</TT> in preparation for a
     * <TT>AASTORE</TT> operation.
     *  
     * Stack: <TT>..., Term[] ref -> ..., Term[] ref, index</TT>
     */
    void arg_store_prep(short index) {
	il.append(new PUSH(cp, index));
    }

    /**
     * Stores the top element of the stack into a <TT>Term[]</TT> array.
     * <P> 
     * Stack: <TT>..., Term[] ref, index, Term -> ...</TT>
     */
    void arg_store_do() {
	il.append(inst_fact.createArrayStore(TermType.JavaType));
    }

    /**
     * Bind a <TT>Binding</TT> to something.
     * <P>
     * Stack: <TT>... -> ...</TT>
     */
    public boolean binding_unify(TermInfo arg0, TermInfo arg1) {
	// we want the stack to look like:
	//  ..., Prolog, Binding (arg0), Term (arg1)
	il.append(JITDefs.prolog_load_inst);
	arg0.load(this, true);
	arg1.load(this, false);

	// take the stack back to: -> ...
	il.append(inst_fact.createInvoke(JITDefs.prolog_class_name,
					 JITDefs.prolog_push,
					 JITDefs.prolog_push_ret_type,
					 JITDefs.prolog_push_args_type,
					 JITDefs.INVOKEVIRTUAL));
	bindings.addElement(arg0);

	return true;
    }

    public void binding_create() {
	il.append(inst_fact.createNew(TypeVariable.JavaType));
	il.append(inst_fact.DUP);
	il.append(inst_fact.ICONST_1);
	il.append(inst_fact.createInvoke(TypeVariable.ClassName,
					 "<init>",
					 Type.VOID,
					 new Type[]{Type.INT},
					 JITDefs.INVOKESPECIAL));
    }

    /**
     * Loads the <TT>args</TT> array of a <TT>Compound</TT> term, stores
     * them in a new LV.
     *
     * Stack: <TT>..., CompoundRef -> ..., Term[] args</TT>
     */
    public LocalVariableGen compound_args_load() {
 	LocalVariableGen lv = lv_alloc(JITDefs.term_array_type);

  	il.append(inst_fact.createGetField(CompoundType.ClassName,
  					   "args",
  					   JITDefs.term_array_type));
	il.append(inst_fact.DUP);
	lv_store(lv);

	return lv;
    }

    /**
     * Compare the <TT>functor_arity</TT> of a <TT>Term</TT> with type
     * <TT>CompoundType</TT> with a given one.
     */
    public void compound_checkFunctor(TermInfo arg0, int functor_arity) {
// 	System.err.println("CCG.compound_checkFunctor(" + arg0.type + ")");

 	BranchInstruction bi;

 	arg0.load(this, true);
 	il.append(inst_fact.createGetField(CompoundType.ClassName,
 					   JITDefs.compound_functor_arity_name,
 					   JITDefs.compound_functor_arity_type));
	il.append(new PUSH(cp, functor_arity));
 	il.append(bi = new IF_ICMPNE(null));
 	addFP(bi);
    }

    /**
     * Creates a new <TT>Compound</TT> object using the argument as a
     * template. Allocates a local variable for the arguments array.
     * <P>
     * Stack: <TT>... -> CompoundRef</TT>
     */
    public void compound_create(TermInfo arg0) {
// 	System.err.println(" >> CCG.compound_create(): " + arg0.type);

	CompoundType ct = (CompoundType)arg0.type;

	if(ct.arity > 0) {
	    il.append(inst_fact.createNew(CompoundType.JavaType));
	    il.append(inst_fact.DUP);

	    String static_field_name = "f" + ct.functor_arity;
	    functors.put(static_field_name, ct.fa);
	    il.append(inst_fact.createGetStatic(class_name,
						static_field_name,
						FunctorArity.JavaType));

	    /* Allocate an LV for the args array. */
	    LocalVariableGen c_args_lv = lv_alloc(JITDefs.term_array_type);

	    arg0.location = new Location(null, c_args_lv);

	    il.append(new PUSH(cp, ct.arity));
	    il.append(new ANEWARRAY(cp.addClass(TermType.JavaType)));
	    il.append(inst_fact.DUP);
	    lv_store(c_args_lv);
	    il.append(inst_fact.createInvoke(CompoundType.ClassName,
					     JITDefs.compound_new,
					     JITDefs.compound_new_ret_type,
					     JITDefs.compound_new_args_type,
					     JITDefs.INVOKESPECIAL));

	    /* By virtue of the normalisation process, all created compound
	     * terms have new variables as arguments. */
	    TermType[] args    = ct.args;
	    TermInfo[] arg_tis = arg0.arg_tis;

	    for(int i = 0; i < arg_tis.length; i++) {
		TermInfo arg_ti = arg_tis[i];

		if(!(arg_ti.type instanceof TypeVariable)) {
		    throw new RuntimeException();
		}

		// Give it a location relative to this new Compound
		// structure.
		if(arg_ti.location != null) {
		    throw new RuntimeException("IS.compound_create(): arg has a location: ["
					       + i + "] " + arg_tis[i]);
		}

		
		arg_ti.location = new Location(arg0, i);
// 		System.err.println(" >> adding arg location: " + arg_ti);
	    }
	} else {
	    String static_field_name = "c" + ct.functor_arity;
	    functors.put(static_field_name, ct.fa);
	    il.append(inst_fact.createGetStatic(class_name,
						static_field_name,
						CompoundType.JavaType));
	}
    }

    /**
     * Assumptions: both have type <TT>CompoundType</TT> with identical
     * functor/arity.
     * <P>
     * Stack: <TT>... -> ...</TT>
     */
    public boolean compound_unify(TermInfo arg0, TermInfo arg1) {
// 	System.err.println("CCG.compound_unify()");

	TermInfo[] arg0_arg_tis = arg0.arg_tis;
	TermInfo[] arg1_arg_tis = arg1.arg_tis;

// 	System.err.println("  arity == " + arg0.arg_tis.length);

	for(int i = 0; i < arg0_arg_tis.length; i++) {
// 	    System.err.println(" >> CCG.compound_unify() arg[" + i + "]\n"
// 			       + arg0_arg_tis[i]
// 			       + "\n <->\n"
// 			       + arg1_arg_tis[i]
// 			       + "\n");

	    if(!arg0_arg_tis[i].type.compile_unify(this,
						   arg0_arg_tis[i],
						   arg1_arg_tis[i])) {
		return false;
	    }
	}

	return true;

//  	throw new RuntimeException();

	/* By virtue of Clause.normalise(), all created compound terms
	 * have new variables as arguments. */
// 	int a = arity;

// 	while(a-- > 0) {
// 	    Location location1 = is.tid.getLocation(args[a]);

// 	    if(location1 == null) {
// 		is.tid.setLocation(args[a], new ArgLocation(cl, a));
// 	    } else {
// 		throw new RuntimeException("Compound.create(): arg has a location: ["
// 					   + a + "] " + args[a]);
    }

    /**
     * Creates a new <TT>FloNum</TT> object.
     */
    public void flonum_create(FloNum n) {
	throw new RuntimeException("ClauseCodeGen.flonum_create(): " + n);
    }

    /**
     * Creates a new <TT>IntNum</TT> object.
     */
    public void intnum_create(IntNum n) {
// 	System.err.println(" >> creating: " + n);
// 	    try {
		// Deal with small-ish int's more efficiently.
		il.append(new PUSH(cp, n.intValue()));
		il.append(inst_fact.createInvoke(JITDefs.intNum_class,
						 JITDefs.intNum_make,
						 JITDefs.intNum_make_ret_type,
						 JITDefs.intNum_make_args_type,
						 JITDefs.INVOKESTATIC));
// 	    } catch(ClassCastException e) {
// 		// OK, it's a big int.
// 		il.append(new PUSH(cp, n.toString(10)));
// 		il.append(inst_fact.createInvoke(JITDefs.intNum_class,
// 						 JITDefs.intNum_valueof,
// 						 JITDefs.intNum_valueof_ret_type,
// 						 JITDefs.intNum_valueof_args_type,
// 						 JITDefs.INVOKESTATIC));
// 	    }
    }

    /**
     * Compare two numbers. Both are assumed to have type
     * <TT>NumericType</TT>.
     *
     * @param mode if compare returns this, success, else failure.
     * @see prolog.math.Numeric#compare
     */
    public boolean numeric_compare(TermInfo arg0, TermInfo arg1, int mode) {
	// FIXME: if they have constants, specialise this. 
	arg0.load(this, true);
	arg1.load(this, true);

	il.append(inst_fact.createInvoke(NumericType.ClassName,
					 JITDefs.numeric_compare,
					 JITDefs.numeric_compare_ret_type,
					 JITDefs.numeric_compare_args_type,
					 JITDefs.INVOKEVIRTUAL));
	il.append(new PUSH(cp, mode));
	addFP(if_icmpne());

	return true;
    }

    /**
     * Unify two numbers. Both are assumed to have type
     * <TT>NumericType</TT>.
     */
    public void numeric_unify(TermInfo arg0, TermInfo arg1) {
	arg0.load(this, true);
	arg1.load(this, true);

	il.append(inst_fact.createInvoke(NumericType.ClassName,
					 "compare",
					 Type.INT,
					 new Type[]{NumericType.JavaType},
					 JITDefs.INVOKEVIRTUAL));
	addFP(ifne());
    }


    public boolean is_not_eq_compile(Term a0, Term a1) {
// 	System.err.println("=\\=.compile():\n\t"
// 			   + a0 + " TI: " + a0.ti
// 			   + ",\n\t"
// 			   + a1 + " TI: " + a1.ti);

	TermInfo arg0 = a0.ti;
	TermInfo arg1 = a1.ti;

	if(arg0.type instanceof NumericType) {
	    arg0.load(this, true);
	} else {
	    compile_eval((Compound)a0);
	}

	if(arg1.type instanceof NumericType) {
	    arg1.load(this, true);
	} else {
	    compile_eval((Compound)a1);
	}

	il.append(inst_fact.createInvoke(NumericType.ClassName,
					 JITDefs.numeric_compare,
					 JITDefs.numeric_compare_ret_type,
					 JITDefs.numeric_compare_args_type,
					 JITDefs.INVOKEVIRTUAL));
	addFP(ifeq());

	return true;
    }

    public boolean is_compile(Term arg0, Term arg1) {
// 	System.err.println("is.compile(): " + arg0 + ", " + arg1);

	/*
	 * BROKEN -- not general
	 * 1. Evaluate RHS.
	 * 2. Compare / bind to LHS.
	 */
	TermInfo ti0 = arg0.ti;
	TermInfo ti1 = arg1.ti;

	// If it's a local var...
	if(ti0.location == null) {
	    // ... allocate new LV, store.
 	    LocalVariableGen lv = lv_alloc(IntType.JavaType);

	    compile_eval((Compound)arg1);
	    lv_store(lv);

	    ti0.type = new IntType();
	    ti0.location = new Location(lv);
	} else {
	    // we want the stack to look like:
	    //  ..., Prolog, Binding (arg0), Term (arg1)
	    il.append(JITDefs.prolog_load_inst);
	    ti0.load(this, true);
	    compile_eval((Compound)arg1);

	    // take the stack back to: -> ...
	    il.append(inst_fact.createInvoke(JITDefs.prolog_class_name,
					     JITDefs.prolog_push,
					     JITDefs.prolog_push_ret_type,
					     JITDefs.prolog_push_args_type,
					     JITDefs.INVOKEVIRTUAL));
	    bindings.addElement(arg0);
	}

	return true;
    }

    /**
     * Leaves a <TT>Numeric</TT> object on the stack.
     */
    public boolean compile_eval(Compound t) {
// 	System.err.println("is_compile_eval(): " + t);

	if(t.arity == 2) {
	    return compile_std_fn(t);
	} else {
	    throw new RuntimeException();
	}
    }

    private boolean compile_std_fn(Compound t) {
	/*
	 * 1. Load LHS, RHS.
	 * 2. Do op.
	 */
	TermInfo ti0 = t.args[0].ti;
	TermInfo ti1 = t.args[1].ti;
	String op = t.functor;

	ti0 = ti0.type.getBinding(ti0);
	ti1 = ti1.type.getBinding(ti1);

	if(ti0.type instanceof NumericType) {
	    ti0.load(this, true);
	} else {
	    // Evaluate it.
	    throw new RuntimeException("compile_std_fn: " + ti0);
	}

	if(ti1.type instanceof NumericType) {
	    ti1.load(this, true);
	} else {
	    // Evaluate it.
	    throw new RuntimeException("compile_std_fn: " + ti1);
	}

	// Do op.
	switch(op.charAt(0)) {
	case '+':
	    il.append(inst_fact.createInvoke(IntType.ClassName,
					     JITDefs.intNum_add,
					     JITDefs.intNum_add_ret_type,
					     JITDefs.intNum_add_args_type,
					     JITDefs.INVOKEVIRTUAL));
	    break;
	case '-':
	    il.append(inst_fact.createInvoke(IntType.ClassName,
					     JITDefs.intNum_sub,
					     JITDefs.intNum_sub_ret_type,
					     JITDefs.intNum_sub_args_type,
					     JITDefs.INVOKEVIRTUAL));
	    break;
	case '/':
	case '%':
	case '*':
// 	    throw new RuntimeException("is/2.compile: " + t);

	default:
	    throw new RuntimeException("is/2: not implemented: " + t);
	}

	return true;
    }


    /**
     * Unify a variable against some other term.
     */
    public boolean variable_unify(TermInfo arg0, TermInfo arg1) {
// 	System.err.println("CCG.variable_unify(): " + arg0 + " <-> " + arg1);

	if(arg0.location == null) {
	    // Allocate a new localVar, load/create arg1, store it.
	    LocalVariableGen lv = lv_alloc(arg1.type.getJavaType(arg1));

// 	    System.err.println("  arg0.loc == null, create/load + store into newLV: " + lv.getName());

	    arg1.load(this, false);
	    lv_store(lv);

	    // Update types, locations, ...
	    arg0.unify(arg1, trail);

	    if(arg1.location == null) {
		arg1.location = new Location(lv);
	    } else {
		arg1.location.set(lv);

// 		throw new RuntimeException();
	    }
	} else {
	    // We're an argument to a Compound. Do a "one-shot"
	    // variable assignment.
// 	    System.err.println("  arg0.loc != null, create/load + store into: " + arg0.location);

	    if(!arg0.location.isArgLocation()) {
		throw new RuntimeException();
	    }

	    arg0.location.store_prep(this);
	    arg1.load(this, false);
	    arg0.location.store_do(this);

	    // Having done the assignment, the var doesn't need a location
	    // anymore.
// 	    arg0.location = null;

	    // Update types, locations, ...
	    arg0.unify(arg1, trail);
	}

	return true;
    }

    /**
     * Unravel binding chains.
     * <P>
     * Stack: <TT>..., TermRef -> (Term)TermRef </TT>
     */
    public void term_getBinding() {
	il.append(inst_fact.createInvoke(TermType.ClassName,
					 JITDefs.term_getBinding,
					 JITDefs.term_getBinding_ret_type,
					 JITDefs.term_getBinding_args_type,
					 JITDefs.INVOKEVIRTUAL));
    }

    /**
     * Unravel binding chains.
     * <P>
     * Stack: <TT>..., TermRef -> (Term)TermRef </TT>
     */
    public void term_getBinding_cast(ObjectType t, InstructionHandle ih) {
	if(t == TermType.JavaType) {
	    throw new RuntimeException();
	}

	il.insert(ih, inst_fact.createInvoke(TermType.ClassName,
					     JITDefs.term_getBinding,
					     JITDefs.term_getBinding_ret_type,
					     JITDefs.term_getBinding_args_type,
					     JITDefs.INVOKEVIRTUAL));
	ih.setInstruction(inst_fact.createCheckCast(t));
    }

    /**
     * Unify two terms together.
     * <P>
     * Stack: <TT>..., TermRef, TermRef -> boolean</TT>
     */
    public void term_unify() {
	il.append(inst_fact.ACONST_NULL);
	il.append(JITDefs.prolog_load_inst);
	il.append(inst_fact.createInvoke(TermType.ClassName,
					 "unify",
					 Type.BOOLEAN,
					 new Type[]{TermType.JavaType,
						    JITDefs.term_array_type,
						    JITDefs.prolog_type},
					 JITDefs.INVOKEVIRTUAL));
	addFP(ifeq());
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append("canFail: ");
	out.append(canFail);
	out.append("\nPredicates used:\n");
	out.append(predicates);
	out.append("\n");
	out.append(il.toString(true));

	return out.toString();
    }
}
