package prolog;

/* The JavaClass package. */
import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;

/**
 * Represents a <TT>Term</TT>'s location, i.e. which LocalVariable it's in.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Location.java,v 1.9 2000/11/16 10:30:08 peteg Exp $
 */
public class Location {
    /**
     * The LV containing this <TT>Term</TT>.
     */
    private LocalVariableGen term_lv;

    /**
     * The LV containing the args array, if loaded.
     */
    private LocalVariableGen args_lv;

    /**
     * The <TT>TermInfo</TT> record for our parent term, if we're an
     * argument to one.
     */
    private TermInfo parent_location;
    private short arg_index;

    /**
     * == <TT>null</TT> ? this term has the Java type dictated by the type
     * field.
     * <P>
     * Otherwise it has type <TT>Term</TT>, but can be cast to something
     * specific by inserting a <TT>checkcast</TT> instruction at this point.
     */
    InstructionHandle cast_pos;

    /**
     * Location of a <TT>Term</TT>.
     *
     * @param term_lv the location of this <TT>Term</TT>.
     */
    Location(LocalVariableGen term_lv) {
	this.term_lv = term_lv;
    }

    /**
     * Location of a <TT>CompoundType</TT>.
     *
     * @param t the location of the term.
     * @param a the location of the args array.
     */
    Location(LocalVariableGen term_lv, LocalVariableGen args_lv) {
	this(term_lv);
	this.args_lv = args_lv;
    }

    /**
     * @param l the location of the <TT>Compound</TT> in which this
     * <TT>Term</TT> is an argument.
     * @param arg_index the position of this argument.
     */
    Location(TermInfo ti, int arg_index) {
	this.parent_location = ti;
	this.arg_index = (short)arg_index;
    }

    void set(LocalVariableGen term_lv) {
	this.term_lv = term_lv;
	this.args_lv = null;
	this.parent_location = null;
    }

    /**
     * Loads this <TT>Term</TT> onto the stack, loading its parent, etc. if
     * necessary.
     *
     * @param cast should this object be cast to it's most-specific type?
     */
    void load(ClauseCodeGen ccg, ObjectType type) {
	if(type != null && cast_pos != null) {
	    // OK, we need something a bit more specific.
// 	    System.err.print(" >> casting: " + cast_pos);
	    ccg.term_getBinding_cast(type, cast_pos);
// 	    System.err.println(" -> " + cast_pos);
	    cast_pos = null;
	}

	if(parent_location != null) {
	    // ArgLocation
// 	    System.err.println("Location.load(Arg): <- " + this);
	    parent_location.location.load_args(ccg);
	    ccg.arg_load(arg_index);

	    if(type != null) {
		// FIXME: the getBinding() might be redundant.
		ccg.term_getBinding();
		ccg.checkcast(type);
	    } else {
		// Put in a NOP that can be later turned into a cast.
		this.cast_pos = ccg.nop();
	    }

	    // FIXME: cache it, turn this into a TermLocation
	    this.term_lv = ccg.lv_alloc(TermType.JavaType); // or ti.type
	    this.parent_location = null;

	    ccg.dup();
	    ccg.lv_store(term_lv);
	} else {
	    // CompoundLocation / TermLocation
// 	    System.err.println("Location.load(Term): <- " + this);
	    ccg.lv_load(term_lv);
	}
    }

    /**
     * Loads a reference to the <TT>Compound</TT>'s args array onto the
     * stack.
     * <P>
     * Stack: <TT>... -> (Term[])ArgsRef</TT>
     */
    void load_args(ClauseCodeGen ccg) {
// 	System.err.println("Location.load_args(Compound): <- " + this);

	if(cast_pos != null) {
	    // Better cast to a Compound first.
// 	    System.err.print(" >> casting: " + cast_pos);
	    ccg.term_getBinding_cast(CompoundType.JavaType, cast_pos);
// 	    ccg.checkcast(CompoundType.JavaType, cast_pos);
// 	    System.err.println(" -> " + cast_pos);
	    cast_pos = null;
	}

	if(args_lv == null) {
	    ccg.lv_load(term_lv);
	    args_lv = ccg.compound_args_load();
// 	    System.err.println("CompoundLocation.load_args(): new args_lv: " + args_lv.getName());
// 	    throw new RuntimeException("CompoundLocation.load_args(): args_lv == null");
	} else {
	    ccg.lv_load(args_lv);
	}
    }

    /**
     * Prepares to store a <TT>Term</TT> into this location. This method is
     * invoked before the <TT>Term</TT> is pushed onto the stack.
     */
    void store_prep(ClauseCodeGen ccg) {
	if(term_lv != null) {
	    throw new RuntimeException("Location.store_prep(Term): " + this);
	} else {
// 	    System.err.println("Location.store_prep(Arg): -> " + this);
	    parent_location.location.load_args(ccg);
	    ccg.arg_store_prep(arg_index);
	}
    }

    /**
     * Actually stores a <TT>Term</TT> into this location. This method is
     * invoked after the <TT>Term</TT> is pushed onto the stack.
     */
    void store_do(ClauseCodeGen ccg) {
	if(term_lv != null) {
// 	    System.err.println("Location.store_do(Term): -> " + this);
	    ccg.lv_store(term_lv);
	} else {
// 	    System.err.println("Location.store_do(Arg): -> " + this);
	    ccg.arg_store_do();
	}
    }

    boolean isArgLocation() {
	return term_lv == null && parent_location != null;
    }

    /**
     * Sets the LV for the <TT>Term</TT> associated with this
     * <TT>Location</TT>.
     */
//     abstract void setTermLV(LocalVariableGen lv);

    public String toString() {
	StringBuffer out = new StringBuffer();

	if(term_lv != null) {
	    out.append("term_lv: ");
	    out.append(term_lv.getName());
	    out.append(" ");
	}

	if(args_lv != null) {
	    out.append("args_lv: ");
	    out.append(args_lv.getName());
	    out.append(" ");
	}

	if(parent_location != null) {
	    out.append("parent_location: (");

	    Location pl = parent_location.location;

	    if(pl == null) {
		out.append("< null >");
	    } else {
		out.append(pl.toString());
	    }

	    out.append(")[");
	    out.append(arg_index);
	    out.append("] ");
	}

	if(cast_pos != null) {
	    out.append("(uncast)");
	}

// 	if(TermLoc) {
// 	    return term_lv.getName();
// 	}

// 	if(CompoundLoc) {
// 	    return "Compound: " + super.toString() + " Args: " +
// 		((args_lv == null) ? "(none)" : args_lv.getName());
// 	}

// 	if(ArgLoc) {
// 	    StringBuffer out = new StringBuffer();

// 	    out.append(parent_location.toString());
// 	    out.append("[");
// 	    out.append(arg_index);
// 	    out.append("]");

// 	    return out.toString();
// 	}

	return out.toString();
    }
}
