package prolog;

import java.io.Serializable;

import prolog.util.Vector;

/**
 * Common features of Prolog terms.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: Term.java,v 1.21 2000/11/03 11:08:10 peteg Exp $
 */
public abstract class Term implements Serializable {
    /**
     * Every <TT>Term</TT> has an associated <TT>TermInfo</TT> (at compile
     * time).
     */
    public TermInfo ti;

    /**
     * Get the <I>Term</I> this <I>Term</I> is bound to. Most <I>Term</I>s
     * are bound to themselves, so to speak.
     */
    public Term getBinding() {
	return this;
    }

    /**
     * Copies this term, instantiating variables from the <CODE>insts</CODE>
     * array. Note this is supposed to do a minimal amount of work as it is
     * used to do lazy structure copying in the interpreter.
     */
    public abstract Term instantiate(Term[] insts);

    /**
     * Unifies this term with the argument. This is the fundamental
     * operation of the resolution process.
     */
    public abstract boolean unify(Term t, Term[] insts, Prolog p);

    /**
     * Asserts that this <TT>Term</TT> is identical to that.
     *
     * @see prolog.lib.identical_2
     * @see prolog.lib.not_identical_2
     */
    public abstract boolean identical(Term t);

    /**
     * Copy a Term, instantiating variables from the array. All variables in
     * the Term must be already used, i.e. no new variables are allocated.
     */
    public abstract Term normalise_copy(Variable[] vars,
					int next_new_var,
					Vector new_vars);

    /**
     * Generate the type tree that describes this <TT>Term</TT>.
     *
     * @arg map map functors to their most general type.
     * @see prolog.TypeDictionary
     */
    public abstract void typify(TypeDictionary types, boolean map);

    /**
     * Returns a displayable version of this object, for use by
     * <TT>write/1</TT>.
     *
     * @see prolog.lib.write_1
     */
    public abstract String displayString();

    /**
     * Returns a representation of this object that may be used for
     * debugging or database listings, etc.
     */
    public abstract String toString();
}
