package prolog;

/* The JavaClass package. */
import de.fub.bytecode.classfile.*;
import de.fub.bytecode.generic.*;

/**
 * Represents a variable that ranges over types. Used to implement
 * polymorphism.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: TypeVariable.java,v 1.13 2000/11/16 10:30:08 peteg Exp $
 */
class TypeVariable extends TermType {
    static final String ClassName    = "prolog.Binding";
    static final ObjectType JavaType = new ObjectType("prolog/Binding");

    private int id;

    /**
     * This <TT>TypeVariable</TT> represents a <TT>Binding</TT> object; so
     * we modify some methods' behaivour accordingly.
     */
    private boolean binding;

    TypeVariable(int id, boolean binding) {
	this.id      = id;
	this.binding = binding;
    }

    ObjectType getJavaType(TermInfo ti) {
	TermInfo binding_ti = ti.typevars[id];

	if(binding_ti == null) {
	    if(binding) {
		return JavaType;
	    } else {
		throw new RuntimeException("TypeVariable.getJavaType(" + this + ")");
	    }
	} else {
	    return binding_ti.type.getJavaType(binding_ti);
	}
    }

    public TermInfo getBinding(TermInfo arg0) {
	TermInfo binding_ti = arg0.typevars[id];

	if(binding_ti == null) {
	    return arg0;
	} else {
	    return binding_ti.type.getBinding(binding_ti);
	}
    }

    TermInfo instantiate(Term t, TermInfo[] typevars,
			 TypeDictionary types, boolean map) {
// 	System.err.println("TypeVariable.instantiate(): " + this + " against " + t);

	if(binding) {
	    throw new RuntimeException();
	}

	if(typevars[id] == null) {
// 	    System.err.println(" >> inst: " + typevars + " [" + id + "]");

	    // Bind to the term's type.
	    t.typify(types, map);

	    return (typevars[id] = t.ti);
	} else {
	    // Let the binding do the work.
// 	    System.err.println(" >> bound: " + typevars + " [" + id + "]");

	    return typevars[id].type.instantiate(t, typevars, types, map);
	}
    }

    boolean unify(TermInfo ti,
		  TermInfo arg,
		  TypeTrail trail) {
// 	System.err.println("TypeVariable.unify(): " + this + " against " + arg.type);

	if(ti.typevars[id] == null) {
	    // FIXME: if we're in a PTD context, then bindings only match
	    // bindings.
	    if(binding) {
		return (arg.type instanceof TypeVariable)
		    && ((TypeVariable)arg.type).binding == true;
	    }

	    ti.typevars[id] = arg;

	    // Trail.
// 	    System.err.println(" >> Trailing: " + this);
	    ti.next     = trail.trail;
	    trail.trail = ti;

	    return true;
	} else {
	    TermInfo binding_ti = ti.typevars[id];

// 	    System.err.println(" >> unify with binding: ");
// 	    System.err.println(binding_ti);

	    return binding_ti.type.unify(binding_ti, arg, trail);
	}
    }

    public boolean compile_unify(ClauseCodeGen ccg,
				 TermInfo arg0, TermInfo arg1) {
	TermInfo binding_ti = arg0.typevars[id];

	if(binding_ti == null) {
 	    return binding
 		? ccg.binding_unify(arg0, arg1)
 		: ccg.variable_unify(arg0, arg1);
	} else {
// 	    System.err.println(" >> TypeVariable.compile_unify(): "
// 			       + arg0 + " <- " + binding_ti);

	    if(arg0.location != null) {
		// FIXME: this is a hack. propogate the location down if
		// possible.
		if(binding_ti.location == null) {
// 		    System.err.println(" >> propogating location down.");
		    binding_ti.location = arg0.location;
		    arg0.location = null;
		} else {
		    throw new RuntimeException(toString());
		}
	    }

	    return binding_ti.type.compile_unify(ccg, binding_ti, arg1);
	}
    }

    public void load(ClauseCodeGen ccg, TermInfo arg0, boolean cast) {
	TermInfo ti = arg0;

// 	System.err.println("TypeVariable.load(): " + this);

	// If we're bound, let it do the work, unless we have a location.
	if(ti.typevars[id] != null) {
	    if(ti.location == null) {
		// OK, we're not doing much.
		TermInfo binding_ti = ti.typevars[id];

		binding_ti.type.load(ccg, binding_ti, cast);
		return;
	    } else {
		throw new RuntimeException("bound to something, location != null");
	    }
	}

	// If we're a binding, and have a location, load as usual.
	if(binding && ti.location != null) {
// 	    System.err.println(" >> binding, loading as usual.");
	    ti.location.load(ccg, cast ? getJavaType(ti) : null);
	    return;
	}

	// If we're not a binding and have no location, create.
	if(!binding && ti.location == null) {
	    create(ccg, ti);
	    return;
	}

	// If we do have a location, it'd better be as an Arg to a Compound
	if(!ti.location.isArgLocation()) {
	    throw new RuntimeException("not an ArgLocation");
	}

	ti.location.store_prep(ccg);
	create(ccg, ti);
	// This is grotty.
	ccg.dup_x2();
	ti.location.store_do(ccg);
    }

    public void create(ClauseCodeGen ccg, TermInfo arg0) {
	TermInfo binding_ti = arg0.typevars[id];

	if(binding_ti == null) {
	    if(binding) {
		throw new RuntimeException(arg0.toString());
	    } else {
		binding = true;

// 		bt.location = new TermLocation(ccg.lv_alloc(BindingType.JavaType));
 		ccg.binding_create();
// 		ccg.dup();
// 		bt.location.store_do(ccg);

// 		System.err.println("TypeVariable.create(): " + this);
	    }
	} else {
	    binding_ti.type.create(ccg, binding_ti);
	}
    }

    public void toString(StringBuffer out) {
	out.append("tv");
	out.append(id);

	if(binding) {
	    out.append(" (binding)");
	}
    }

    void unbind(TermInfo ti) {
	if(ti.typevars[id] == null) {
	    throw new RuntimeException();
	} else {
// 	    System.err.println(" >> Unbinding: " + this);
	    ti.typevars[id] = null;
	    ti.location = null;
	    ti.next = null;
	}
    }

    void setID(int id) {
	this.id = id;
    }
}
