package prolog.lib;

import prolog.*;
import prolog.math.*;

/**
 * <TT>is/2</TT>: Arithmetic evaluation. Success occurs iff the first
 * argument is unifiable with the result of arithmetically evaluating the
 * second (which is assumed to be a valid expression on pain of a
 * <TT>ArgumentException</TT>).
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: is_2.java,v 1.18 2000/11/13 11:55:04 peteg Exp $
 * @see prolog.lib.is_equal_2
 * @see prolog.lib.is_not_equal_2
 */
public class is_2 extends CompiledPredicate {
    public static final String functor = "is".intern();

    public is_2() {
    }

    public boolean exec(Term[] args, Prolog p) throws ArgumentException {
// 	System.err.println(args[0] + " is " + args[1]);

	try {
	    Term arg0 = args[0].getBinding();
	    Term arg1 = args[1].getBinding();

	    if(arg1 instanceof Compound) {
		return arg0.unify(eval((Compound)arg1), null, p);
	    } else if(arg1 instanceof Numeric) {
		return arg0.unify(arg1, null, p);
	    }
	} catch(ClassCastException e) {
	    // Fall through to the generic error handler.
	    throw new RuntimeException(e.toString());
	}

 	throw new ArgumentException("is/2: invalid arguments: ", args);
    }

    /**
     * Evaluates an expression tree. We let other classes use this method to
     * evaluate their arguments.
     */
    public static Numeric eval(Compound t) throws ArgumentException {
	if(t.arity == 2) {
	    // Standard binary operators.
	    Term t0 = t.args[0].getBinding();
	    Term t1 = t.args[1].getBinding();
	    Numeric v0, v1;
	    String op = t.functor;

	    if(t0 instanceof Compound) {
		v0 = eval((Compound)t0);
	    } else {
		v0 = (Numeric)t0;
	    }

	    if(t1 instanceof Compound) {
		v1 = eval((Compound)t1);
	    } else {
		v1 = (Numeric)t1;
	    }

	    switch(op.charAt(0)) {
	    case '+':
		return v0.add(v1);
	    case '/':
		if(op.length() > 1) {
		    /* Integer division only works on int's. */
		    try {
			Numeric i0 = v0;
			Numeric i1 = v1;
// 		IntNum i0 = (IntNum)v0;
// 		IntNum i1 = (IntNum)v1;

//  			System.err.println("is/2: //: i0 = " + i0 + " i1 = " + i1 + " i0 // i1 = " + i0.div(i1).toInt(Numeric.TRUNCATE));

			return i0.div(i1).toInt(Numeric.TRUNCATE);
		    } catch (ClassCastException e) {
			throw new ArgumentException("int div operator applied to non-integral operands: ", t.args);
		    }
		} else {
		    return v0.div(v1);
		}
	    case '%': /* Modulo only works on int's. */
//  	    System.err.println("is/2: %: v0 = " + v0 + " v1 = " + v1 + " v0 % v1 = " + v0.mod(v1));

		return v0.mod(v1);
// 	    try {
// 		IntNum i0 = (IntNum)v0;
// 		IntNum i1 = (IntNum)v1;

// 		return IntNum.modulo(i0, i1);
// 	    } catch (ClassCastException e) {
// 		throw new ArgumentException("mod operator applied to non-integral operands: ", t.args);
// 	    }
	    case '*':
		return v0.mul(v1);
	    case '-':
		return v0.sub(v1);

	    default:
		throw new ArgumentException("is/2: invalid operation: ", t);
	    }
	} else {
	    // Transcendental functions.
	    Term t0 = t.args[0].getBinding();
	    Numeric v0;
	    String op = t.functor;

	    if(t0 instanceof Compound) {
		v0 = (Numeric)eval((Compound)t0);
	    } else {
		v0 = (Numeric)t0;
	    }

	    if(op.equals("abs"))	return v0.abs();
	    if(op.equals("asin"))	return v0.asin();
// 	    if(op.equals("cos"))	return v0.cos();
	    if(op.equals("exp"))	return v0.exp();
	    if(op.equals("log"))	return v0.log();
// 	    if(op.equals("random"))	return v0.sin();
	    if(op.equals("sin"))	return v0.sin();
	    if(op.equals("sqrt"))	return v0.sqrt();

	    throw new ArgumentException("is/2: invalid operation: ", t);
	}
    }

    public boolean compile_call(ClauseCodeGen ccg, Term[] args, boolean lastcall) {
	return ccg.is_compile(args[0], args[1]);
    }
}
