package prolog.lib;

import prolog.*;

/**
 * <TT>retract/1</TT>: Removes a clause from a predicate. 
 * <P>
 * <TT>retract(pred_name(x1, x2, ..., xn))</TT> or
 * <TT>retract(':-'(Head, Body))</TT>.
 * <P>
 * Note that pred_name must be bound to a <TT>Compound</TT> (although it's
 * arguments may be unbound). The argument is unified with the retracted
 * clause.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: retract_1.java,v 1.14 2000/08/22 12:11:39 peteg Exp $
 * @see prolog.lib.abolish_1
 * @see prolog.lib.asserta_1
 * @see prolog.lib.assertz_1
 */
public class retract_1 extends Predicate {
    public retract_1() {
    }

    public boolean unify(Term[] args, Prolog p) throws ArgumentException {
//    	System.err.println("retract: " + args[0]);

	Compound arg0;
	String functor;
	int arity;
	DynamicPredicate predicate;

	try {
	    arg0 = (Compound)args[0].getBinding();

	    if(arg0.functor.equals(":-") && arg0.arity == 2) {
		functor = ((Compound)arg0.args[0]).functor;
		arity = ((Compound)arg0.args[0]).arity;
	    } else {
		functor = arg0.functor;
		arity = arg0.arity;
	    }
	    predicate = (DynamicPredicate)p.db.get(functor, arity);
	} catch (ClassCastException e) {
	    throw new ArgumentException("retract/1: invalid argument: ", args);
	}

	if(predicate == null) {
// 	    System.err.println("retract/1: predicate not defined: \""
// 			       + functor + "/" + arity + "\"");

	    return false;
	}

	return retract(predicate, arg0, predicate.getClauses(), p);
    }

    public boolean redo(Term[] args, Prolog p, ChoicePoint cp)
	throws ArgumentException {

	return retract(cp.predicate, cp.arg0, cp.next_clause, p);
    }

    private boolean retract(DynamicPredicate predicate, Compound arg0,
			    ClauseList cl, Prolog p) throws ArgumentException {
	Binding trail_top = p.get_trail_top();

	/* Find a clause that unifies with arg0. */
	while(cl != null) {
	    Clause c = cl.clause;
	    Compound parsed = c.parsed;
	    Term[] insts = new Term[c.numVars];

	    if(parsed.unify(arg0, insts, p)) {
//  		System.err.println("retract - succeed: " + c);

		predicate.remove(c);

		if(cl.next != null) {
		    /* Have further clauses to examine. */
		    ChoicePoint cp = p.allocate_cp(cl.next, trail_top);

		    cp.arg0      = arg0;
		    cp.predicate = predicate;
		}

		p.use_pending_goal();

		return true;
	    } else {
//  		System.err.println("retract - skipping: " + c);
		p.unwind_trail(trail_top);
		cl = cl.next;
	    }
	}

//  	System.err.println("retract - failed.");

	return false;
    }

    public String toString() {
	return "\n(builtin)\n";
    }
}
