%IPROLOG

% Model checker for a toy programming language
% Author Harsha Perera



check(Filename, Input, Output) :- 
       parse(Variables, Initial, Final, Program, Filename),
       !,
       generate(Input, Variables),
       run(Output, Initial, Final, Program, Input).
 
	
check(_, syntaxerror, syntaxerror).


% TOKENIZING THE PROGRAM FILE

get_tokens(File,L):- 
       see(File), 
       tokens(L), 
       seen(File). 

tokens(R) :- 
       ratom(A),!, 
       (A = end_of_file -> 
       R=[] | R= [A|R1], tokens(R1)). 

	
parse(Variables, Initial, Final, Program, File) :-
	get_tokens(File,L),
	varPhrase(Variables, L, L1),
	initPhrase(Initial, L1, Variables, [';'|L2]),
	finalPhrase(Final, L2, Variables, [';'|L3]),
	L3=['begin'|_],
	program(Program, L3, Variables, []).
	
% PARSING AND IDENTIFYING VARIABLES
	
varPhrase(Variables, ['variables'|L], L1) :-
	varList(L, Variables, L1).
	
varList([X,','|L], [X|Y], L1) :-
	validVar(X), 
	varList(L, Y, L1).
	
varList([X,';'|L], [X], L) :-
	validVar(X).
	
% Eliminate invalid variable names
	
validVar('begin') :- !, fail.
		
validVar('end') :- !, fail.

validVar('if') :- !, fail.

validVar('then') :- !, fail.

validVar('else') :- !, fail.

validVar('variables') :- !, fail.

validVar('initialcondition') :- !, fail.

validVar('finalcondition') :- !, fail.

validVar('not') :- !, fail.

validVar('while'):-!,fail.

validVar('do'):-!,fail.

		
	
validVar(X):-
	atom_chars(X,[Y|Z]),
	char_code(Y,Q),
	char_code('a',P),
	char_code('z',R),
	P=<Q,
	Q=<R.

% PARSING THE INITIAL AND FINAL CONDITIONS

initPhrase(Initial, ['initialcondition'|L], Variables, L1) :-
	condition(Initial, L, Variables, L1).

finalPhrase(Final, ['finalcondition'|L], Variables, L1) :-
	condition(Final, L, Variables, L1).
	
% PARSING A CONDITION
	
condition(Cond, [Cond|L], Variables, L):- member(Cond, Variables).

condition(conj(Cond1,Cond2), L1, Variables, L6):-
	L1=['('|L2],
	condition(Cond1, L2, Variables, L3),
	L3=['/\'|L4],
	condition(Cond2, L4, Variables, L5),
	L5=[')'|L6].
	
condition(disj(Cond1,Cond2), L1, Variables, L6):-
	L1=['('|L2],
	condition(Cond1, L2, Variables, L3),
	L3=['\/'|L4],
	condition(Cond2, L4, Variables, L5),
	L5=[')'|L6].

condition(impl(Cond1,Cond2), L1, Variables, L6):-
	L1=['('|L2],
	condition(Cond1, L2, Variables, L3),
	L3=['->'|L4],
	condition(Cond2, L4, Variables, L5),
	L5=[')'|L6].
	
condition(negt(Cond), L1, Variables, L4):-
	L1=['not','('|L2],
	condition(Cond, L2, Variables, L3),
	L3=[')'|L4].
	
% PARSING THE PROGRAM INSTRUCTIONS

% A program is considered to be a list of statements and each statement itself 
% could possibly contain a list of statements. Therefore the main program is
% a singleton list containing a block statement.

program([],[],_,[]).

program([], L1, _, L1):- 
	L1=['end'|L2].

program([X|R],L1,Variables,L3):-
	statement(X,L1,Variables,L2),
	program(R,L2,Variables,L3).
		
statement(block(X), L1, Variables, L4):-
	L1=['begin'|L2],
	program(X, L2, Variables, L3),
	L3=['end'|L4].

statement(assgt(Var,Cond), L1, Variables, L3):-
	L1=[Var,':='|L2],
	member(Var,Variables),
	condition(Cond, L2, Variables, [';','else'|L3]),!,fail.
	
statement(assgt(Var,Cond), L1, Variables, L3):-
	L1=[Var,':='|L2],
	member(Var,Variables),
	condition(Cond, L2, Variables, [';','end'|L3]),!,fail.
	
		
statement(assgt(Var,Cond), L1, Variables, L3):-
	L1=[Var,':='|L2],
	member(Var,Variables),
	condition(Cond, L2, Variables, [';'|L3]),!.
	
statement(assgt(Var,Cond), L1, Variables, L3):-
	L1=[Var,':='|L2],
	member(Var,Variables),
	condition(Cond, L2, Variables, L3).
	
	
statement(if(Cond,[Block1],[Block2]), L1, Variables, L7):-
        L1=['if'|L2],
        condition(Cond, L2, Variables, L3),
        L3=['then'|L4],
        statement(Block1, L4, Variables, L5),
        L5=['else'|L6],
        statement(Block2, L6, Variables, L7).
	
statement(while(Cond,[Block]),L1,Variables,L5):-
        L1=['while'|L2],
	condition(Cond,L2,Variables,L3),
	L3=['do'|L4],
	statement(Block,L4,Variables,L5).
		
        
 
        		 								       
% GENERATING VARIOUS INPUT TO SEARCH FOR A COUNTER EXAMPLE
       
generate([],[]).
generate([val(X,Y)|Z],[X|R]):-
	value(Y),
	generate(Z,R).
	
value(true).
value(false).

% CHECKING WHETHER AN INPUT IS VALID AND CAN BE USED TO GENERATE 
% 			COUNTER EXAMPLE

run(Output, Initial, Final, Program, Input):-
	eval(X,Input,Initial),
	X=true,
	execute(Output,Input,Program),
	eval(Y,Output,Final),
	Y=false.
	
	
% EVALUATING THE VALUES OF CONDITIONS

eval(false,Assignment,_):-
	Assignment=nonterminating,!.

eval(true,Assignment,conj(Cond1,Cond2)):-
	eval(X,Assignment,Cond1),
	X=true,
        eval(Y,Assignment,Cond2),
	Y=true,
	!.	
		
eval(false,Assignment,conj(_,_)).
	
eval(true,Assignment,disj(Cond1,Cond2)):-
	eval(X,Assignment,Cond1),
	X=true,!.

eval(true,Assignment,disj(Cond1,Cond2)):-
	eval(X,Assignment,Cond2),
	X=true,!.
	
eval(false,Assignment,disj(_,_)).

eval(true,Assignment,negt(Cond)):-
	eval(X,Assignment,Cond),
	X=false,!.
	
eval(false,Assignment,negt(Cond)).

eval(true,Assignment,impl(Cond1,Cond2)):-
	eval(X,Assignment,Cond1),
	X=false,!.
	
eval(true,Assignment,impl(Cond1,Cond2)):-
	eval(X,Assignment,Cond2),
	X=true,!.
	
eval(false,Assignment,impl(_,_)).

eval(Value,Assignment,X):-
	valueof(Value,Assignment,X).
	
valueof(Value,[val(X,Value)|_],X).

valueof(Value,[_|R],X):-	
	valueof(Value,R,X).
	
	
% EXECUTING PROGRAM STATEMENTS

execute(Input,Input,[]).

execute(Output,Input,[block(X)|R]):-
	execute(Output1,Input,X),
	execute(Output,Output1,R).
	
execute(Output,Input,[assgt(Var,Cond)|R]):-
	eval(Value,Input,Cond),
	swap(Output1,Input,val(Var,Value)),
	execute(Output,Output1,R).
	
execute(Output,Input,[if(Cond,Block1,Block2)|R]):-
	eval(X,Input,Cond),
	X=true,
	execute(Output1,Input,Block1),
	execute(Output,Output1,R).
	
execute(Output,Input,[if(Cond,Block1,Block2)|R]):-
	eval(X,Input,Cond),
	X=false,
	execute(Output1,Input,Block2),
	execute(Output,Output1,R).

% Executing while loops
% Check for nontermination - If a program has n elements the maximum number of iterations for a while loop 
% is considered to be the number of different states that the variables can be in. (Two to the power of n)
% A loop attempting to execute beyond this point is considered to be one that does not terminate.
 	
execute(Output,Input,[while(Cond,Block)|R]):-
	elts(X,Input),
	maxexecs(Max,X),
	runwhile(Output,Input,[while(Cond,Block)|R],Max,0).

runwhile(Output,Input,[while(Cond,Block)|R],Max,Iteration):-
	eval(X,Input,Cond),
	X=false,
	execute(Output,Input,R).
	
runwhile(nonterminating,Input,[while(Cond,Block)|R],Max,Iteration):-
	Iteration>Max,!.
	
runwhile(Output,Input,[while(Cond,Block)|R],Max,Iteration):-
	eval(X,Input,Cond),
	X=true,
	execute(Output1,Input,Block),
	Nextiteration is Iteration+1,
	runwhile(Output,Output1,[while(Cond,Block)|R],Max,Nextiteration).		


% HELPER PREDICATES FOR EXECUTING PROGRAMS

swap(Output,[val(Var,_)|R],val(Var,Value)):-
	Output=[val(Var,Value)|R],!.
	
	
swap(Output,[X|R],val(Var,Value)):-
	Output=[X|Output1],
	swap(Output1,R,val(Var,Value)),!.

	
maxexecs(X,0):-
	X=1,!.
	
maxexecs(X,Num):-
	 Y is Num-1,
         maxexecs(Z,Y),
	 X is 2*Z.
		
elts(0,[]).
elts(M,[X|Y]):-elts(N,Y),M is N+1.	
			



	
	
	
	

	
	

 	
	  
          
