package prolog;

import prolog.util.Hashtable;

/**
 * Represents closed <TT>Lists</TT> (of the form <TT>[x1, x2, ..., xn | []
 * ]</TT>), where the cdr of the last cell is the <TT>EmptyList</TT>.
 *
 * @author Peter Gammie (<A HREF="MailTo:peteg@cse.unsw.edu.au">peteg@cse.unsw.edu.au</A>)
 * @version $Id: List.java,v 1.1 2000/08/06 13:43:46 peteg Exp $
 */
public class List extends Compound {
    /**
     * Represents the list functor.
     */
    public final static String listFunctor = ".".intern();

    /**
     * Creates a <TT>List</TT> element with given <TT>car</TT> and
     * <TT>cdr</TT>.
     */
    public List(Term car, Term cdr) {
	super(listFunctor, new Term[]{car, cdr});
    }

    public List(Term[] args) {
	super(listFunctor, args);
    }

    /**
     * For the <TT>EmptyList</TT>.
     */
    protected List(String functor) {
	super(functor);
    }

    public boolean unify(Term t, Term[] insts, Prolog p) {
// 	System.out.println("Trying to unify: " + this + " with: " + t);

	if(t instanceof Variable) {
	    return t.unify(this, insts, p);
	}

	if(!(t instanceof List)) {
// 	    System.err.println("List.unify(): failed unify.\n\t" + this
// 			       + "\n\t" + t);
	    return false;
	}

	Term[] as = args;
	Term[] bs = ((Compound)t).args;

	return as[0].unify(bs[0], insts, p) && as[1].unify(bs[1], insts, p);
    }

    public boolean identical(Term t) {
	if(!(t instanceof List))
	    return(false);

	Compound c = (Compound)t;

	return(args[0].identical(c.args[0]) && args[1].identical(c.args[1]));
    }

    protected Compound squash(Hashtable vars) {
 	Term [] as = new Term[arity];
 	Variable v;

	for(int i = 2; i-- > 0; ) {
	    as[i] = args[i].getBinding();

	    if(as[i] instanceof Compound) {
		as[i] = ((Compound)as[i]).squash(vars);
	    } else {
		/* If it's a Binding, change it into a Variable. */
		if(as[i] instanceof Binding) {
		    if((v = (Variable)vars.get(as[i])) == null)
			vars.put(as[i], v = new Variable(vars.size()));
		    as[i] = v;
		}
	    }
	}

	return(new List(as));
    }

    public String displayString() {
	return toString();
    }

    public String toString() {
	StringBuffer out = new StringBuffer();

	out.append("[");

	Compound cur = this;

	while(true) {
	    Term t0 = cur.args[0].getBinding();
	    Term t1 = cur.args[1].getBinding();

	    out.append(t0.toString());

	    if(t1 instanceof Compound) {
		Compound next = (Compound)t1;

		if(next.functor == listFunctor && next.arity == 2) {
		    out.append(", ");
		    cur = next;
		    continue;
		} else if(next == EmptyList.emptyList) {
		    break;
		}
	    }

	    out.append("|");
	    out.append(t1.toString());
	    break;
	}

	out.append("]");
	return out.toString();
    }
}
